from __future__ import annotations

from importlib import import_module
from importlib.util import find_spec


def is_valid_import(module: str, /, *, name: str | None = None) -> bool:
    """Check if an import is valid."""
    spec = find_spec(module)
    if spec is None:
        return False
    if name is None:
        return True
    mod = import_module(module)
    return hasattr(mod, name)


__all__ = ["is_valid_import"]



from __future__ import annotations

from operator import add, eq, ge, gt, le, lt, mul, ne, sub, truediv
from re import search
from typing import TYPE_CHECKING, Any

from pytest import mark, param, raises
from whenever import TimeDelta

from utilities.asyncio import sleep_td
from utilities.timer import Timer
from utilities.whenever import SECOND, ZERO_TIME

if TYPE_CHECKING:
    from collections.abc import Callable


class TestTimer:
    @mark.parametrize(
        ("op", "other"),
        [
            param(add, ZERO_TIME),
            param(sub, ZERO_TIME),
            param(mul, 1),
            param(mul, 1.0),
            param(truediv, 1),
            param(truediv, 1.0),
        ],
    )
    def test_arithmetic_against_numbers_or_timedeltas(
        self, *, op: Callable[[Any, Any], Any], other: Any
    ) -> None:
        with Timer() as timer:
            pass
        assert isinstance(op(timer, other), TimeDelta)

    @mark.parametrize(
        ("op", "cls"),
        [param(add, TimeDelta), param(sub, TimeDelta), param(truediv, float)],
    )
    async def test_arithmetic_against_another_timer(
        self, *, op: Callable[[Any, Any], Any], cls: type[Any]
    ) -> None:
        with Timer() as timer1, Timer() as timer2:
            await sleep_td(0.01 * SECOND)
        assert isinstance(op(timer1, timer2), cls)

    @mark.parametrize(("op"), [param(add), param(sub), param(mul), param(truediv)])
    def test_arithmetic_error(self, *, op: Callable[[Any, Any], Any]) -> None:
        with Timer() as timer:
            pass
        with raises(TypeError):
            _ = op(timer, "")

    @mark.parametrize(
        ("op", "expected"),
        [
            param(eq, False),
            param(ne, True),
            param(ge, False),
            param(gt, False),
            param(le, True),
            param(lt, True),
        ],
    )
    def test_comparison(
        self, *, op: Callable[[Any, Any], bool], expected: bool
    ) -> None:
        with Timer() as timer:
            pass
        assert op(timer, SECOND) is expected

    @mark.parametrize(
        "op", [param(eq), param(ne), param(ge), param(gt), param(le), param(lt)]
    )
    def test_comparison_between_timers(self, *, op: Callable[[Any, Any], bool]) -> None:
        with Timer() as timer1:
            pass
        with Timer() as timer2:
            pass
        assert isinstance(op(timer1, timer2), bool)

    @mark.parametrize(("op", "expected"), [param(eq, False), param(ne, True)])
    def test_comparison_eq_and_ne(
        self, *, op: Callable[[Any, Any], bool], expected: bool
    ) -> None:
        with Timer() as timer:
            pass
        assert op(timer, "") is expected

    @mark.parametrize("op", [param(ge), param(gt), param(le), param(lt)])
    def test_comparison_error(self, *, op: Callable[[Any, Any], bool]) -> None:
        with Timer() as timer:
            pass
        with raises(TypeError):
            _ = op(timer, "")

    async def test_context_manager(self) -> None:
        delta = 0.1 * SECOND
        with Timer() as timer:
            await sleep_td(2 * delta)
        assert timer >= delta

    def test_float(self) -> None:
        with Timer() as timer:
            ...
        assert float(timer) == timer.timedelta.in_seconds()

    def test_hashable(self) -> None:
        timer = Timer()
        _ = hash(timer)

    @mark.parametrize("func", [param(repr), param(str)])
    async def test_repr_and_str(self, *, func: Callable[[Timer], str]) -> None:
        with Timer() as timer:
            await sleep_td(0.01 * SECOND)
        as_str = func(timer)
        assert search(r"^PT0\.\d+S$", as_str)

    async def test_running(self) -> None:
        delta = 0.1 * SECOND
        timer = Timer()
        await sleep_td(2 * delta)
        assert timer >= delta
        await sleep_td(2 * delta)
        assert timer >= 2 * delta

    def test_timedelta(self) -> None:
        timer = Timer()
        assert isinstance(timer.timedelta, TimeDelta)



from __future__ import annotations

from zoneinfo import ZoneInfo

from utilities.types import TimeZone
from utilities.typing import get_literal_elements
from utilities.tzlocal import LOCAL_TIME_ZONE, LOCAL_TIME_ZONE_NAME, get_local_time_zone


class TestGetLocalTimeZone:
    def test_function(self) -> None:
        time_zone = get_local_time_zone()
        assert isinstance(time_zone, ZoneInfo)

    def test_constants(self) -> None:
        assert isinstance(LOCAL_TIME_ZONE, ZoneInfo)
        assert LOCAL_TIME_ZONE_NAME in get_literal_elements(TimeZone)



from __future__ import annotations

from typing import TYPE_CHECKING

from hypothesis import HealthCheck, given, settings
from hypothesis.strategies import integers

from utilities.hypothesis import text_ascii
from utilities.shelve import yield_shelf

if TYPE_CHECKING:
    from pathlib import Path


class TestYieldShelf:
    @given(key=text_ascii(), value=integers())
    @settings(
        max_examples=1, suppress_health_check={HealthCheck.function_scoped_fixture}
    )
    def test_main(self, *, key: str, value: int, tmp_path: Path) -> None:
        path = tmp_path.joinpath("shelf")
        with yield_shelf(path) as shelf:
            shelf[key] = value
        with yield_shelf(path) as shelf:
            result = shelf[key]
        assert result == value



from __future__ import annotations

from asyncio import Queue
from itertools import chain, repeat
from typing import TYPE_CHECKING, Any

from hypothesis import HealthCheck, Phase, given, settings
from hypothesis.strategies import (
    DataObject,
    binary,
    booleans,
    data,
    dictionaries,
    lists,
    permutations,
    sampled_from,
)
from pytest import mark, param, raises
from redis.asyncio import Redis
from redis.asyncio.client import PubSub

from tests.conftest import SKIPIF_CI_AND_NOT_LINUX
from tests.test_objects.objects import objects
from utilities.asyncio import get_items_nowait, sleep_td
from utilities.hypothesis import int64s, pairs, text_ascii
from utilities.operator import is_equal
from utilities.orjson import deserialize, serialize
from utilities.redis import (
    PublishError,
    _is_message,
    _RedisMessage,
    publish,
    publish_many,
    redis_hash_map_key,
    redis_key,
    subscribe,
    yield_pubsub,
    yield_redis,
)
from utilities.sentinel import SENTINEL_REPR, Sentinel, sentinel
from utilities.text import unique_str
from utilities.whenever import MICROSECOND, SECOND

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

    from whenever import TimeDelta


_PUB_SUB_SLEEP: TimeDelta = 0.1 * SECOND


class TestIsMessage:
    @mark.parametrize(
        ("message", "channels", "expected"),
        [
            param(
                {
                    "type": "message",
                    "pattern": None,
                    "channel": b"channel",
                    "data": b"data",
                },
                [b"channel"],
                True,
            ),
            param(None, [], False),
            param({"type": "invalid"}, [], False),
            param({"type": "message"}, [], False),
            param({"type": "message", "pattern": False}, [], False),
            param({"type": "message", "pattern": None}, [], False),
            param(
                {"type": "message", "pattern": None, "channel": b"channel1"},
                [b"channel2"],
                False,
            ),
            param(
                {"type": "message", "pattern": None, "channel": b"channel"},
                [b"channel"],
                False,
            ),
            param(
                {
                    "type": "message",
                    "pattern": None,
                    "channel": b"channel",
                    "data": None,
                },
                [b"channel"],
                False,
            ),
        ],
    )
    def test_main(
        self, *, message: Any, channels: Sequence[bytes], expected: bool
    ) -> None:
        result = _is_message(message, channels=channels)
        assert result is expected


class TestPublish:
    @given(data=lists(binary(min_size=1), min_size=1))
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_bytes(self, *, test_redis: Redis, data: Sequence[bytes]) -> None:
        channel = unique_str()
        queue: Queue[bytes] = Queue()
        async with subscribe(test_redis, channel, queue, output="bytes"):
            await sleep_td(_PUB_SUB_SLEEP)
            for datum in data:
                _ = await publish(test_redis, channel, datum)
            await sleep_td(_PUB_SUB_SLEEP)  # remain in context
        assert queue.qsize() == len(data)
        results = get_items_nowait(queue)
        for result, datum in zip(results, data, strict=True):
            assert isinstance(result, bytes)
            assert result == datum

    @given(objects=lists(objects(), min_size=1))
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_serializer(
        self, *, test_redis: Redis, objects: Sequence[Any]
    ) -> None:
        channel = unique_str()
        queue: Queue[Any] = Queue()
        async with subscribe(test_redis, channel, queue, output=deserialize):
            await sleep_td(_PUB_SUB_SLEEP)
            for obj in objects:
                _ = await publish(test_redis, channel, obj, serializer=serialize)
            await sleep_td(_PUB_SUB_SLEEP)  # remain in context
        assert queue.qsize() == len(objects)
        results = get_items_nowait(queue)
        for result, obj in zip(results, objects, strict=True):
            assert is_equal(result, obj)

    @given(messages=lists(text_ascii(min_size=1), min_size=1))
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_text(self, *, test_redis: Redis, messages: Sequence[str]) -> None:
        channel = f"test_{unique_str()}"
        queue: Queue[str] = Queue()
        async with subscribe(test_redis, channel, queue):
            await sleep_td(_PUB_SUB_SLEEP)
            for message in messages:
                _ = await publish(test_redis, channel, message)
            await sleep_td(_PUB_SUB_SLEEP)  # remain in context
        assert queue.qsize() == len(messages)
        results = get_items_nowait(queue)
        for result, message in zip(results, messages, strict=True):
            assert isinstance(result, str)
            assert result == message

    async def test_error(self, *, test_redis: Redis) -> None:
        with raises(
            PublishError, match="Unable to publish data None with no serializer"
        ):
            _ = await publish(test_redis, "channel", None)


class TestPublishMany:
    @given(
        data=lists(binary(min_size=1) | text_ascii(min_size=1) | objects(), min_size=1)
    )
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_main(self, *, test_redis: Redis, data: Sequence[Any]) -> None:
        result = await publish_many(
            test_redis, unique_str(), data, serializer=serialize
        )
        expected = list(repeat(object=True, times=len(data)))
        assert result == expected

    @given(messages=lists(text_ascii(min_size=1), min_size=1))
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_timeout(self, *, test_redis: Redis, messages: Sequence[str]) -> None:
        result = await publish_many(
            test_redis, unique_str(), messages, timeout=MICROSECOND
        )
        expected = list(repeat(object=False, times=len(messages)))
        assert result == expected


class TestRedisHashMapKey:
    @given(key=int64s(), value=booleans())
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_get_and_set_bool(
        self, *, test_redis: Redis, key: int, value: bool
    ) -> None:
        hm_key = redis_hash_map_key(unique_str(), int, bool)
        _ = await hm_key.set(test_redis, key, value)
        assert await hm_key.get(test_redis, key) is value

    @given(key=booleans() | int64s(), value=booleans())
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_get_and_set_union_key(
        self, *, test_redis: Redis, key: bool | int, value: bool
    ) -> None:
        hm_key = redis_hash_map_key(unique_str(), (bool, int), bool)
        _ = await hm_key.set(test_redis, key, value)
        assert await hm_key.get(test_redis, key) is value

    @given(value=booleans())
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_get_and_set_sentinel_key(
        self, *, test_redis: Redis, value: bool
    ) -> None:
        def serializer(sentinel: Sentinel, /) -> bytes:
            return repr(sentinel).encode()

        hm_key = redis_hash_map_key(
            unique_str(), Sentinel, bool, key_serializer=serializer
        )
        _ = await hm_key.set(test_redis, sentinel, value)
        assert await hm_key.get(test_redis, sentinel) is value

    @given(key=int64s(), value=int64s() | booleans())
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_get_and_set_union_value(
        self, *, test_redis: Redis, key: int, value: bool | int
    ) -> None:
        hm_key = redis_hash_map_key(unique_str(), int, (bool, int))
        _ = await hm_key.set(test_redis, key, value)
        assert await hm_key.get(test_redis, key) == value

    @given(key=int64s())
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_get_and_set_sentinel_value(
        self, *, test_redis: Redis, key: int
    ) -> None:
        def serializer(sentinel: Sentinel, /) -> bytes:
            return repr(sentinel).encode()

        def deserializer(data: bytes, /) -> Sentinel:
            assert data == SENTINEL_REPR.encode()
            return sentinel

        hm_key = redis_hash_map_key(
            unique_str(),
            int,
            Sentinel,
            value_serializer=serializer,
            value_deserializer=deserializer,
        )
        _ = await hm_key.set(test_redis, key, sentinel)
        assert await hm_key.get(test_redis, key) is sentinel

    @given(data=data(), mapping=dictionaries(int64s(), booleans()))
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_get_and_set_many(
        self, *, test_redis: Redis, data: DataObject, mapping: Mapping[int, bool]
    ) -> None:
        hm_key = redis_hash_map_key(unique_str(), int, bool)
        _ = await hm_key.set_many(test_redis, mapping)
        if len(mapping) == 0:
            keys = []
        else:
            keys = data.draw(lists(sampled_from(list(mapping))))
        expected = [mapping[k] for k in keys]
        assert await hm_key.get_many(test_redis, keys) == expected

    @given(key=int64s(), value=booleans())
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_delete(self, *, test_redis: Redis, key: int, value: bool) -> None:
        hm_key = redis_hash_map_key(unique_str(), int, bool)
        _ = await hm_key.set(test_redis, key, value)
        assert await hm_key.get(test_redis, key) is value
        _ = await hm_key.delete(test_redis, key)
        with raises(KeyError):
            _ = await hm_key.get(test_redis, key)

    @given(key=pairs(int64s()), value=booleans())
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_delete_compound(
        self, *, test_redis: Redis, key: tuple[int, int], value: bool
    ) -> None:
        hm_key = redis_hash_map_key(unique_str(), tuple[int, int], bool)
        _ = await hm_key.set(test_redis, key, value)
        assert await hm_key.get(test_redis, key) is value
        _ = await hm_key.delete(test_redis, key)
        with raises(KeyError):
            _ = await hm_key.get(test_redis, key)

    @given(key=int64s(), value=booleans())
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_exists(self, *, test_redis: Redis, key: int, value: bool) -> None:
        hm_key = redis_hash_map_key(unique_str(), int, bool)
        assert not (await hm_key.exists(test_redis, key))
        _ = await hm_key.set(test_redis, key, value)
        assert await hm_key.exists(test_redis, key)

    @given(key=pairs(int64s()), value=booleans())
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_exists_compound(
        self, *, test_redis: Redis, key: tuple[int, int], value: bool
    ) -> None:
        hm_key = redis_hash_map_key(unique_str(), tuple[int, int], bool)
        assert not (await hm_key.exists(test_redis, key))
        _ = await hm_key.set(test_redis, key, value)
        assert await hm_key.exists(test_redis, key)

    @given(mapping=dictionaries(int64s(), booleans()))
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_get_all(
        self, *, test_redis: Redis, mapping: Mapping[int, bool]
    ) -> None:
        hm_key = redis_hash_map_key(unique_str(), int, bool)
        _ = await hm_key.set_many(test_redis, mapping)
        assert await hm_key.get_all(test_redis) == mapping

    @given(mapping=dictionaries(int64s(), booleans()))
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_keys(
        self, *, test_redis: Redis, mapping: Mapping[int, bool]
    ) -> None:
        hm_key = redis_hash_map_key(unique_str(), int, bool)
        _ = await hm_key.set_many(test_redis, mapping)
        assert await hm_key.keys(test_redis) == list(mapping)

    @given(mapping=dictionaries(int64s(), booleans()))
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_length(
        self, *, test_redis: Redis, mapping: Mapping[int, bool]
    ) -> None:
        hm_key = redis_hash_map_key(unique_str(), int, bool)
        _ = await hm_key.set_many(test_redis, mapping)
        assert await hm_key.length(test_redis) == len(mapping)

    @given(key=int64s(), value=booleans())
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_ttl(self, *, test_redis: Redis, key: int, value: bool) -> None:
        delta = 0.1 * SECOND
        hm_key = redis_hash_map_key(unique_str(), int, bool, ttl=2 * delta)
        _ = await hm_key.set(test_redis, key, value)
        await sleep_td(delta)  # else next line may not work
        assert await hm_key.exists(test_redis, key)
        await sleep_td(2 * delta)
        assert not await test_redis.exists(hm_key.name)

    @given(mapping=dictionaries(int64s(), booleans()))
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_values(
        self, *, test_redis: Redis, mapping: Mapping[int, bool]
    ) -> None:
        hm_key = redis_hash_map_key(unique_str(), int, bool)
        _ = await hm_key.set_many(test_redis, mapping)
        assert await hm_key.values(test_redis) == list(mapping.values())


class TestRedisKey:
    @given(value=booleans())
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_get_and_set_bool(self, *, test_redis: Redis, value: bool) -> None:
        key = redis_key(unique_str(), bool)
        _ = await key.set(test_redis, value)
        assert await key.get(test_redis) is value

    @given(value=booleans() | int64s())
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_get_and_set_union(
        self, *, test_redis: Redis, value: bool | int
    ) -> None:
        key = redis_key(unique_str(), (bool, int))
        _ = await key.set(test_redis, value)
        assert await key.get(test_redis) == value

    @mark.flaky
    async def test_get_and_set_sentinel_with_serialize(
        self, *, test_redis: Redis
    ) -> None:
        def serializer(sentinel: Sentinel, /) -> bytes:
            return repr(sentinel).encode()

        def deserializer(data: bytes, /) -> Sentinel:
            assert data == SENTINEL_REPR.encode()
            return sentinel

        red_key = redis_key(
            unique_str(), Sentinel, serializer=serializer, deserializer=deserializer
        )
        _ = await red_key.set(test_redis, sentinel)
        assert await red_key.get(test_redis) is sentinel

    @given(value=booleans())
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_delete(self, *, test_redis: Redis, value: bool) -> None:
        key = redis_key(unique_str(), bool)
        _ = await key.set(test_redis, value)
        assert await key.get(test_redis) is value
        _ = await key.delete(test_redis)
        with raises(KeyError):
            _ = await key.get(test_redis)

    @given(value=booleans())
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_exists(self, *, test_redis: Redis, value: bool) -> None:
        key = redis_key(unique_str(), bool)
        assert not (await key.exists(test_redis))
        _ = await key.set(test_redis, value)
        assert await key.exists(test_redis)

    @given(value=booleans())
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_ttl(self, *, test_redis: Redis, value: bool) -> None:
        delta = 0.1 * SECOND
        key = redis_key(unique_str(), bool, ttl=2 * delta)
        _ = await key.set(test_redis, value)
        await sleep_td(delta)  # else next line may not work
        assert await key.exists(test_redis)
        await sleep_td(2 * delta)
        assert not await key.exists(test_redis)


class TestSubscribe:
    @given(messages=lists(binary(min_size=1), min_size=1))
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_bytes(self, *, test_redis: Redis, messages: Sequence[bytes]) -> None:
        channel = unique_str()
        queue: Queue[bytes] = Queue()
        async with subscribe(test_redis, channel, queue, output="bytes"):
            await sleep_td(_PUB_SUB_SLEEP)
            for message in messages:
                await test_redis.publish(channel, message)
            await sleep_td(_PUB_SUB_SLEEP)  # remain in context
        assert queue.qsize() == len(messages)
        results = get_items_nowait(queue)
        for result, message in zip(results, messages, strict=True):
            assert isinstance(result, bytes)
            assert result == message

    @given(objs=lists(objects(), min_size=1))
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_deserialize(self, *, test_redis: Redis, objs: Sequence[Any]) -> None:
        channel = unique_str()
        queue: Queue[Any] = Queue()
        async with subscribe(test_redis, channel, queue, output=deserialize):
            await sleep_td(_PUB_SUB_SLEEP)
            for obj in objs:
                await test_redis.publish(channel, serialize(obj))
            await sleep_td(_PUB_SUB_SLEEP)  # remain in context
        assert queue.qsize() == len(objs)
        results = get_items_nowait(queue)
        for result, obj in zip(results, objs, strict=True):
            assert is_equal(result, obj)

    @given(
        data=data(),
        short_messages=lists(text_ascii(max_size=4), min_size=1),
        long_messages=lists(text_ascii(min_size=6), min_size=1),
    )
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_filter(
        self,
        *,
        test_redis: Redis,
        data: DataObject,
        short_messages: Sequence[str],
        long_messages: Sequence[str],
    ) -> None:
        channel = unique_str()
        messages = data.draw(permutations(list(chain(short_messages, long_messages))))
        queue: Queue[str] = Queue()
        async with subscribe(
            test_redis, channel, queue, filter_=lambda text: len(text) >= 6
        ):
            await sleep_td(_PUB_SUB_SLEEP)
            for message in messages:
                await test_redis.publish(channel, message)
            await sleep_td(_PUB_SUB_SLEEP)  # remain in context
        assert queue.qsize() == len(long_messages)
        results = get_items_nowait(queue)
        for result in results:
            assert isinstance(result, str)
            assert len(result) >= 3

    @given(messages=lists(text_ascii(min_size=1), min_size=1))
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_raw(self, *, test_redis: Redis, messages: Sequence[str]) -> None:
        channel = f"test_{unique_str()}"
        queue: Queue[_RedisMessage] = Queue()
        async with subscribe(test_redis, channel, queue, output="raw"):
            await sleep_td(_PUB_SUB_SLEEP)
            for message in messages:
                await test_redis.publish(channel, message)
            await sleep_td(_PUB_SUB_SLEEP)  # remain in context
        assert queue.qsize() == len(messages)
        results = get_items_nowait(queue)
        for result, message in zip(results, messages, strict=True):
            assert isinstance(result, dict)
            assert result["type"] == "message"
            assert result["pattern"] is None
            assert result["channel"] == channel.encode()
            assert result["data"] == message.encode()

    @given(messages=lists(text_ascii(min_size=1), min_size=1))
    @mark.flaky
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_text(self, *, test_redis: Redis, messages: Sequence[str]) -> None:
        channel = f"test_{unique_str()}"
        queue: Queue[_RedisMessage] = Queue()
        async with subscribe(test_redis, channel, queue, output="raw"):
            await sleep_td(_PUB_SUB_SLEEP)
            for message in messages:
                await test_redis.publish(channel, message)
            await sleep_td(_PUB_SUB_SLEEP)  # remain in context
        assert queue.qsize() == len(messages)
        results = get_items_nowait(queue)
        for result, message in zip(results, messages, strict=True):
            assert isinstance(result, dict)
            assert result["type"] == "message"
            assert result["pattern"] is None
            assert result["channel"] == channel.encode()
            assert result["data"] == message.encode()


class TestYieldRedis:
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_main(self) -> None:
        async with yield_redis() as client:
            assert isinstance(client, Redis)


class TestYieldPubSub:
    async def test_main(self, *, test_redis: Redis) -> None:
        async with yield_pubsub(test_redis, unique_str()) as pubsub:
            assert isinstance(pubsub, PubSub)



from __future__ import annotations

import datetime as dt
from typing import TYPE_CHECKING, Literal, cast
from zoneinfo import ZoneInfo

from hypothesis import given
from hypothesis.strategies import DataObject, data, datetimes, just, sampled_from
from pytest import mark, param, raises

from utilities.hypothesis import zone_infos, zoned_date_times
from utilities.tzdata import HongKong, Tokyo
from utilities.tzlocal import LOCAL_TIME_ZONE, LOCAL_TIME_ZONE_NAME
from utilities.zoneinfo import (
    UTC,
    _ToTimeZoneNameInvalidKeyError,
    _ToTimeZoneNameInvalidTZInfoError,
    _ToTimeZoneNamePlainDateTimeError,
    _ToZoneInfoInvalidTZInfoError,
    _ToZoneInfoPlainDateTimeError,
    to_time_zone_name,
    to_zone_info,
)

if TYPE_CHECKING:
    from utilities.types import TimeZoneLike


class TestToZoneInfo:
    @given(time_zone=zone_infos())
    def test_zone_info(self, *, time_zone: ZoneInfo) -> None:
        result = to_zone_info(time_zone)
        assert result is time_zone

    @given(data=data(), time_zone=zone_infos())
    def test_zoned_date_time(self, *, data: DataObject, time_zone: ZoneInfo) -> None:
        date_time = data.draw(zoned_date_times(time_zone=time_zone))
        result = to_zone_info(date_time)
        assert result is time_zone

    @mark.parametrize("time_zone", [param("local"), param("localtime")])
    def test_local(self, *, time_zone: Literal["local", "localtime"]) -> None:
        result = to_zone_info(time_zone)
        assert result is LOCAL_TIME_ZONE

    @given(time_zone=zone_infos())
    def test_str(self, *, time_zone: ZoneInfo) -> None:
        result = to_zone_info(cast("TimeZoneLike", time_zone.key))
        assert result is time_zone

    def test_tz_info(self) -> None:
        result = to_zone_info(dt.UTC)
        assert result is UTC

    @given(data=data(), time_zone=zone_infos())
    def test_py_zoned_date_time(self, *, data: DataObject, time_zone: ZoneInfo) -> None:
        date_time = data.draw(datetimes(timezones=just(time_zone)))
        result = to_zone_info(date_time)
        assert result is time_zone

    def test_error_invalid_tz_info(self) -> None:
        time_zone = dt.timezone(dt.timedelta(hours=12))
        with raises(
            _ToZoneInfoInvalidTZInfoError, match=r"Invalid time-zone: UTC\+12:00"
        ):
            _ = to_zone_info(time_zone)

    @given(date_time=datetimes())
    def test_error_plain_date_time(self, *, date_time: dt.datetime) -> None:
        with raises(_ToZoneInfoPlainDateTimeError, match="Plain date-time: .*"):
            _ = to_zone_info(date_time)


class TestToTimeZoneName:
    @given(time_zone=zone_infos())
    def test_zone_info(self, *, time_zone: ZoneInfo) -> None:
        result = to_time_zone_name(time_zone)
        expected = time_zone.key
        assert result == expected

    @given(data=data(), time_zone=zone_infos())
    def test_zoned_date_time(self, *, data: DataObject, time_zone: ZoneInfo) -> None:
        date_time = data.draw(zoned_date_times(time_zone=time_zone))
        result = to_time_zone_name(date_time)
        expected = time_zone.key
        assert result == expected

    @mark.parametrize("time_zone", [param("local"), param("localtime")])
    def test_local(self, *, time_zone: Literal["local", "localtime"]) -> None:
        result = to_time_zone_name(time_zone)
        assert result == LOCAL_TIME_ZONE_NAME

    @given(time_zone=zone_infos())
    def test_str(self, *, time_zone: ZoneInfo) -> None:
        result = to_time_zone_name(cast("TimeZoneLike", time_zone.key))
        expected = time_zone.key
        assert result == expected

    def test_tz_info(self) -> None:
        result = to_time_zone_name(dt.UTC)
        expected = UTC.key
        assert result == expected

    @given(data=data(), time_zone=zone_infos())
    def test_py_zoned_date_time(self, *, data: DataObject, time_zone: ZoneInfo) -> None:
        date_time = data.draw(datetimes(timezones=just(time_zone)))
        result = to_time_zone_name(date_time)
        expected = time_zone.key
        assert result == expected

    def test_error_invalid_key(self) -> None:
        with raises(
            _ToTimeZoneNameInvalidKeyError, match="Invalid time-zone: 'invalid'"
        ):
            _ = to_time_zone_name(cast("TimeZoneLike", "invalid"))

    def test_error_invalid_tz_info(self) -> None:
        time_zone = dt.timezone(dt.timedelta(hours=12))
        with raises(_ToTimeZoneNameInvalidTZInfoError, match="Invalid time-zone: .*"):
            _ = to_time_zone_name(time_zone)

    @given(date_time=datetimes())
    def test_error_plain_date_time(self, *, date_time: dt.datetime) -> None:
        with raises(_ToTimeZoneNamePlainDateTimeError, match="Plain date-time: .*"):
            _ = to_time_zone_name(date_time)


class TestTimeZones:
    @given(time_zone=sampled_from([HongKong, Tokyo, UTC]))
    def test_main(self, *, time_zone: ZoneInfo) -> None:
        assert isinstance(time_zone, ZoneInfo)



from __future__ import annotations

from asyncio import CancelledError, Queue, Task, TaskGroup, create_task
from collections.abc import AsyncIterator, Callable, Mapping, Sequence
from contextlib import suppress
from dataclasses import dataclass
from functools import partial
from operator import itemgetter
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    TypedDict,
    TypeGuard,
    assert_never,
    cast,
    overload,
    override,
)

from redis.asyncio import Redis

from utilities.asyncio import sleep_td, timeout_td
from utilities.contextlib import enhanced_async_context_manager
from utilities.errors import ImpossibleCaseError
from utilities.functions import ensure_int, identity
from utilities.iterables import always_iterable, one
from utilities.os import is_pytest
from utilities.whenever import MILLISECOND, SECOND, to_milliseconds, to_seconds

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Awaitable, Collection, Iterable

    from redis.asyncio import ConnectionPool
    from redis.asyncio.client import PubSub
    from redis.typing import EncodableT

    from utilities.iterables import MaybeIterable
    from utilities.types import Delta, MaybeSequence, MaybeType, TypeLike


_PUBLISH_TIMEOUT: Delta = SECOND


##


@dataclass(kw_only=True)
class RedisHashMapKey[K, V]:
    """A hashmap key in a redis store."""

    name: str
    key: TypeLike[K]
    key_serializer: Callable[[K], bytes] | None = None
    key_deserializer: Callable[[bytes], K] | None = None
    value: TypeLike[V]
    value_serializer: Callable[[V], bytes] | None = None
    value_deserializer: Callable[[bytes], V] | None = None
    timeout: Delta | None = None
    error: MaybeType[BaseException] = TimeoutError
    ttl: Delta | None = None

    async def delete(self, redis: Redis, key: K, /) -> int:
        """Delete a key from a hashmap in `redis`."""
        ser = _serialize(  # skipif-ci-and-not-linux
            key, serializer=self.key_serializer
        ).decode()
        async with timeout_td(  # skipif-ci-and-not-linux
            self.timeout, error=self.error
        ):
            return await cast("Awaitable[int]", redis.hdel(self.name, ser))
        raise ImpossibleCaseError(case=[f"{redis=}", f"{key=}"])  # pragma: no cover

    async def exists(self, redis: Redis, key: K, /) -> bool:
        """Check if the key exists in a hashmap in `redis`."""
        ser = _serialize(  # skipif-ci-and-not-linux
            key, serializer=self.key_serializer
        ).decode()
        async with timeout_td(  # skipif-ci-and-not-linux
            self.timeout, error=self.error
        ):
            return await cast("Awaitable[bool]", redis.hexists(self.name, ser))

    async def get(self, redis: Redis, key: K, /) -> V:
        """Get a value from a hashmap in `redis`."""
        result = one(await self.get_many(redis, [key]))  # skipif-ci-and-not-linux
        if result is None:  # skipif-ci-and-not-linux
            raise KeyError(self.name, key)
        return result  # skipif-ci-and-not-linux

    async def get_all(self, redis: Redis, /) -> Mapping[K, V]:
        """Get a value from a hashmap in `redis`."""
        async with timeout_td(  # skipif-ci-and-not-linux
            self.timeout, error=self.error
        ):
            result = await cast(  # skipif-ci-and-not-linux
                "Awaitable[Mapping[bytes, bytes]]", redis.hgetall(self.name)
            )
        return {  # skipif-ci-and-not-linux
            _deserialize(key, deserializer=self.key_deserializer): _deserialize(
                value, deserializer=self.value_deserializer
            )
            for key, value in result.items()
        }

    async def get_many(self, redis: Redis, keys: Iterable[K], /) -> Sequence[V | None]:
        """Get multiple values from a hashmap in `redis`."""
        keys = list(keys)  # skipif-ci-and-not-linux
        if len(keys) == 0:  # skipif-ci-and-not-linux
            return []
        ser = [  # skipif-ci-and-not-linux
            _serialize(key, serializer=self.key_serializer) for key in keys
        ]
        async with timeout_td(  # skipif-ci-and-not-linux
            self.timeout, error=self.error
        ):
            result = await cast(  # skipif-ci-and-not-linux
                "Awaitable[Sequence[bytes | None]]", redis.hmget(self.name, ser)
            )
        return [  # skipif-ci-and-not-linux
            None
            if data is None
            else _deserialize(data, deserializer=self.value_deserializer)
            for data in result
        ]

    async def keys(self, redis: Redis, /) -> Sequence[K]:
        """Get the keys of a hashmap in `redis`."""
        async with timeout_td(  # skipif-ci-and-not-linux
            self.timeout, error=self.error
        ):
            result = await cast("Awaitable[Sequence[bytes]]", redis.hkeys(self.name))
        return [  # skipif-ci-and-not-linux
            _deserialize(data, deserializer=self.key_deserializer) for data in result
        ]

    async def length(self, redis: Redis, /) -> int:
        """Get the length of a hashmap in `redis`."""
        async with timeout_td(  # skipif-ci-and-not-linux
            self.timeout, error=self.error
        ):
            return await cast("Awaitable[int]", redis.hlen(self.name))

    async def set(self, redis: Redis, key: K, value: V, /) -> int:
        """Set a value in a hashmap in `redis`."""
        return await self.set_many(redis, {key: value})  # skipif-ci-and-not-linux

    async def set_many(self, redis: Redis, mapping: Mapping[K, V], /) -> int:
        """Set multiple value(s) in a hashmap in `redis`."""
        if len(mapping) == 0:  # skipif-ci-and-not-linux
            return 0
        ser = {  # skipif-ci-and-not-linux
            _serialize(key, serializer=self.key_serializer): _serialize(
                value, serializer=self.value_serializer
            )
            for key, value in mapping.items()
        }
        async with timeout_td(  # skipif-ci-and-not-linux
            self.timeout, error=self.error
        ):
            result = await cast(
                "Awaitable[int]", redis.hset(self.name, mapping=cast("Any", ser))
            )
            if self.ttl is not None:
                await redis.pexpire(self.name, to_milliseconds(self.ttl))
        return result  # skipif-ci-and-not-linux

    async def values(self, redis: Redis, /) -> Sequence[V]:
        """Get the values of a hashmap in `redis`."""
        async with timeout_td(  # skipif-ci-and-not-linux
            self.timeout, error=self.error
        ):
            result = await cast("Awaitable[Sequence[bytes]]", redis.hvals(self.name))
        return [  # skipif-ci-and-not-linux
            _deserialize(data, deserializer=self.value_deserializer) for data in result
        ]


@overload
def redis_hash_map_key[K, V](
    name: str,
    key: type[K],
    value: type[V],
    /,
    *,
    key_serializer: Callable[[K], bytes] | None = None,
    key_deserializer: Callable[[bytes], Any] | None = None,
    value_serializer: Callable[[V], bytes] | None = None,
    value_deserializer: Callable[[bytes], V] | None = None,
    timeout: Delta | None = None,
    error: MaybeType[BaseException] = TimeoutError,
    ttl: Delta | None = None,
) -> RedisHashMapKey[K, V]: ...
@overload
def redis_hash_map_key[K, V1, V2](
    name: str,
    key: type[K],
    value: tuple[type[V1], type[V2]],
    /,
    *,
    key_serializer: Callable[[K], bytes] | None = None,
    key_deserializer: Callable[[bytes], Any] | None = None,
    value_serializer: Callable[[V1 | V2], bytes] | None = None,
    value_deserializer: Callable[[bytes], V1 | V2] | None = None,
    timeout: Delta | None = None,
    error: MaybeType[BaseException] = TimeoutError,
    ttl: Delta | None = None,
) -> RedisHashMapKey[K, V1 | V2]: ...
@overload
def redis_hash_map_key[K, V1, V2, V3](
    name: str,
    key: type[K],
    value: tuple[type[V1], type[V2], type[V3]],
    /,
    *,
    key_serializer: Callable[[K], bytes] | None = None,
    key_deserializer: Callable[[bytes], Any] | None = None,
    value_serializer: Callable[[V1 | V2 | V3], bytes] | None = None,
    value_deserializer: Callable[[bytes], V1 | V2 | V3] | None = None,
    timeout: Delta | None = None,
    error: MaybeType[BaseException] = TimeoutError,
    ttl: Delta | None = None,
) -> RedisHashMapKey[K, V1 | V2 | V3]: ...
@overload
def redis_hash_map_key[K1, K2, V](
    name: str,
    key: tuple[type[K1], type[K2]],
    value: type[V],
    /,
    *,
    key_serializer: Callable[[K1 | K2], bytes] | None = None,
    key_deserializer: Callable[[bytes], Any] | None = None,
    value_serializer: Callable[[V], bytes] | None = None,
    value_deserializer: Callable[[bytes], V] | None = None,
    timeout: Delta | None = None,
    error: MaybeType[BaseException] = TimeoutError,
    ttl: Delta | None = None,
) -> RedisHashMapKey[K1 | K2, V]: ...
@overload
def redis_hash_map_key[K1, K2, V1, V2](
    name: str,
    key: tuple[type[K1], type[K2]],
    value: tuple[type[V1], type[V2]],
    /,
    *,
    key_serializer: Callable[[K1 | K2], bytes] | None = None,
    key_deserializer: Callable[[bytes], Any] | None = None,
    value_serializer: Callable[[V1 | V2], bytes] | None = None,
    value_deserializer: Callable[[bytes], V1 | V2] | None = None,
    timeout: Delta | None = None,
    error: MaybeType[BaseException] = TimeoutError,
    ttl: Delta | None = None,
) -> RedisHashMapKey[K1 | K2, V1 | V2]: ...
@overload
def redis_hash_map_key[K1, K2, V1, V2, V3](
    name: str,
    key: tuple[type[K1], type[K2]],
    value: tuple[type[V1], type[V2], type[V3]],
    /,
    *,
    key_serializer: Callable[[K1 | K2], bytes] | None = None,
    key_deserializer: Callable[[bytes], Any] | None = None,
    value_serializer: Callable[[V1 | V2 | V3], bytes] | None = None,
    value_deserializer: Callable[[bytes], V1 | V2 | V3] | None = None,
    timeout: Delta | None = None,
    error: MaybeType[BaseException] = TimeoutError,
    ttl: Delta | None = None,
) -> RedisHashMapKey[K1 | K2, V1 | V2 | V3]: ...
@overload
def redis_hash_map_key[K1, K2, K3, V](
    name: str,
    key: tuple[type[K1], type[K2], type[K3]],
    value: type[V],
    /,
    *,
    key_serializer: Callable[[K1 | K2 | K3], bytes] | None = None,
    key_deserializer: Callable[[bytes], Any] | None = None,
    value_serializer: Callable[[V], bytes] | None = None,
    value_deserializer: Callable[[bytes], V] | None = None,
    timeout: Delta | None = None,
    error: MaybeType[BaseException] = TimeoutError,
    ttl: Delta | None = None,
) -> RedisHashMapKey[K1 | K2 | K3, V]: ...
@overload
def redis_hash_map_key[K1, K2, K3, V1, V2](
    name: str,
    key: tuple[type[K1], type[K2], type[K3]],
    value: tuple[type[V1], type[V2]],
    /,
    *,
    key_serializer: Callable[[K1 | K2 | K3], bytes] | None = None,
    key_deserializer: Callable[[bytes], Any] | None = None,
    value_serializer: Callable[[V1 | V2], bytes] | None = None,
    value_deserializer: Callable[[bytes], V1 | V2] | None = None,
    timeout: Delta | None = None,
    error: MaybeType[BaseException] = TimeoutError,
    ttl: Delta | None = None,
) -> RedisHashMapKey[K1 | K2 | K3, V1 | V2]: ...
@overload
def redis_hash_map_key[K1, K2, K3, V1, V2, V3](
    name: str,
    key: tuple[type[K1], type[K2], type[K3]],
    value: tuple[type[V1], type[V2], type[V3]],
    /,
    *,
    key_serializer: Callable[[K1 | K2 | K3], bytes] | None = None,
    key_deserializer: Callable[[bytes], Any] | None = None,
    value_serializer: Callable[[V1 | V2 | V3], bytes] | None = None,
    value_deserializer: Callable[[bytes], V1 | V2 | V3] | None = None,
    timeout: Delta | None = None,
    error: MaybeType[BaseException] = TimeoutError,
    ttl: Delta | None = None,
) -> RedisHashMapKey[K1 | K2 | K3, V1 | V2 | V3]: ...
@overload
def redis_hash_map_key[K, K1, K2, K3, V, V1, V2, V3](
    name: str,
    key: TypeLike[K],
    value: TypeLike[V],
    /,
    *,
    key_serializer: Callable[[K1 | K2 | K3], bytes] | None = None,
    key_deserializer: Callable[[bytes], Any] | None = None,
    value_serializer: Callable[[V1 | V2 | V3], bytes] | None = None,
    value_deserializer: Callable[[bytes], V1 | V2 | V3] | None = None,
    timeout: Delta | None = None,
    error: MaybeType[BaseException] = TimeoutError,
    ttl: Delta | None = None,
) -> RedisHashMapKey[K, V]: ...
def redis_hash_map_key[K, V](
    name: str,
    key: TypeLike[K],
    value: TypeLike[V],
    /,
    *,
    key_serializer: Callable[[Any], bytes] | None = None,
    key_deserializer: Callable[[bytes], Any] | None = None,
    value_serializer: Callable[[Any], bytes] | None = None,
    value_deserializer: Callable[[bytes], Any] | None = None,
    timeout: Delta | None = None,
    ttl: Delta | None = None,
    error: MaybeType[BaseException] = TimeoutError,
) -> RedisHashMapKey[K, V]:
    """Create a redis key."""
    return RedisHashMapKey(  # skipif-ci-and-not-linux
        name=name,
        key=key,
        key_serializer=key_serializer,
        key_deserializer=key_deserializer,
        value=value,
        value_serializer=value_serializer,
        value_deserializer=value_deserializer,
        timeout=timeout,
        error=error,
        ttl=ttl,
    )


##


@dataclass(kw_only=True)
class RedisKey[T]:
    """A key in a redis store."""

    name: str
    type: TypeLike[T]
    serializer: Callable[[T], bytes] | None = None
    deserializer: Callable[[bytes], T] | None = None
    timeout: Delta | None = None
    error: MaybeType[BaseException] = TimeoutError
    ttl: Delta | None = None

    async def delete(self, redis: Redis, /) -> int:
        """Delete the key from `redis`."""
        async with timeout_td(  # skipif-ci-and-not-linux
            self.timeout, error=self.error
        ):
            response = await redis.delete(self.name)
        return ensure_int(response)

    async def exists(self, redis: Redis, /) -> bool:
        """Check if the key exists in `redis`."""
        async with timeout_td(  # skipif-ci-and-not-linux
            self.timeout, error=self.error
        ):
            result = cast("Literal[0, 1]", await redis.exists(self.name))
        match result:  # skipif-ci-and-not-linux
            case 0 | 1 as value:
                return bool(value)
            case never:
                assert_never(never)

    async def get(self, redis: Redis, /) -> T:
        """Get a value from `redis`."""
        async with timeout_td(  # skipif-ci-and-not-linux
            self.timeout, error=self.error
        ):
            result = cast("bytes | None", await redis.get(self.name))
        if result is None:  # skipif-ci-and-not-linux
            raise KeyError(self.name)
        return _deserialize(  # skipif-ci-and-not-linux
            result, deserializer=self.deserializer
        )

    async def set(self, redis: Redis, value: T, /) -> int:
        """Set a value in `redis`."""
        ser = _serialize(value, serializer=self.serializer)  # skipif-ci-and-not-linux
        ttl = (  # skipif-ci-and-not-linux
            None if self.ttl is None else to_milliseconds(self.ttl)
        )
        async with timeout_td(  # skipif-ci-and-not-linux
            self.timeout, error=self.error
        ):
            response = await redis.set(  # skipif-ci-and-not-linux
                self.name, ser, px=ttl
            )
        return ensure_int(response)  # skipif-ci-and-not-linux


@overload
def redis_key[T](
    name: str,
    type_: type[T],
    /,
    *,
    serializer: Callable[[T], bytes] | None = None,
    deserializer: Callable[[bytes], T] | None = None,
    timeout: Delta | None = None,
    error: MaybeType[BaseException] = TimeoutError,
    ttl: Delta | None = None,
) -> RedisKey[T]: ...
@overload
def redis_key[T1, T2](
    name: str,
    type_: tuple[type[T1], type[T2]],
    /,
    *,
    serializer: Callable[[T1 | T2], bytes] | None = None,
    deserializer: Callable[[bytes], T1 | T2] | None = None,
    timeout: Delta | None = None,
    error: MaybeType[BaseException] = TimeoutError,
    ttl: Delta | None = None,
) -> RedisKey[T1 | T2]: ...
@overload
def redis_key[T1, T2, T3](
    name: str,
    type_: tuple[type[T1], type[T2], type[T3]],
    /,
    *,
    serializer: Callable[[T1 | T2 | T3], bytes] | None = None,
    deserializer: Callable[[bytes], T1 | T2 | T3] | None = None,
    timeout: Delta | None = None,
    error: MaybeType[BaseException] = TimeoutError,
    ttl: Delta | None = None,
) -> RedisKey[T1 | T2 | T3]: ...
@overload
def redis_key[T1, T2, T3, T4](
    name: str,
    type_: tuple[type[T1], type[T2], type[T3], type[T4]],
    /,
    *,
    serializer: Callable[[T1 | T2 | T3 | T4], bytes] | None = None,
    deserializer: Callable[[bytes], T1 | T2 | T3 | T4] | None = None,
    timeout: Delta | None = None,
    error: MaybeType[BaseException] = TimeoutError,
    ttl: Delta | None = None,
) -> RedisKey[T1 | T2 | T3 | T4]: ...
@overload
def redis_key[T1, T2, T3, T4, T5](
    name: str,
    type_: tuple[type[T1], type[T2], type[T3], type[T4], type[T5]],
    /,
    *,
    serializer: Callable[[T1 | T2 | T3 | T4 | T5], bytes] | None = None,
    deserializer: Callable[[bytes], T1 | T2 | T3 | T4 | T5] | None = None,
    timeout: Delta | None = None,
    error: MaybeType[BaseException] = TimeoutError,
    ttl: Delta | None = None,
) -> RedisKey[T1 | T2 | T3 | T4 | T5]: ...
@overload
def redis_key[T, T1, T2, T3, T4, T5](
    name: str,
    type_: TypeLike[T],
    /,
    *,
    serializer: Callable[[T1 | T2 | T3 | T4 | T5], bytes] | None = None,
    deserializer: Callable[[bytes], T1 | T2 | T3 | T4 | T5] | None = None,
    timeout: Delta | None = None,
    error: MaybeType[BaseException] = TimeoutError,
    ttl: Delta | None = None,
) -> RedisKey[T]: ...
def redis_key[T](
    name: str,
    type_: TypeLike[T],
    /,
    *,
    serializer: Callable[[Any], bytes] | None = None,
    deserializer: Callable[[bytes], Any] | None = None,
    timeout: Delta | None = None,
    error: MaybeType[BaseException] = TimeoutError,
    ttl: Delta | None = None,
) -> RedisKey[T]:
    """Create a redis key."""
    return RedisKey(  # skipif-ci-and-not-linux
        name=name,
        type=type_,
        serializer=serializer,
        deserializer=deserializer,
        timeout=timeout,
        error=error,
        ttl=ttl,
    )


##


@overload
async def publish[T](
    redis: Redis,
    channel: str,
    data: T,
    /,
    *,
    serializer: Callable[[T], EncodableT],
    timeout: Delta | None = _PUBLISH_TIMEOUT,
) -> int: ...
@overload
async def publish(
    redis: Redis,
    channel: str,
    data: bytes | str,
    /,
    *,
    serializer: None = None,
    timeout: Delta | None = _PUBLISH_TIMEOUT,
) -> int: ...
@overload
async def publish[T](
    redis: Redis,
    channel: str,
    data: bytes | str | T,
    /,
    *,
    serializer: Callable[[T], EncodableT] | None = None,
    timeout: Delta | None = _PUBLISH_TIMEOUT,
) -> int: ...
async def publish[T](
    redis: Redis,
    channel: str,
    data: bytes | str | T,
    /,
    *,
    serializer: Callable[[T], EncodableT] | None = None,
    timeout: Delta | None = _PUBLISH_TIMEOUT,
) -> int:
    """Publish an object to a channel."""
    match data, serializer:  # skipif-ci-and-not-linux
        case bytes() | str() as data_use, _:
            ...
        case _, None:
            raise PublishError(data=data)
        case _, Callable():
            data_use = serializer(data)
        case never:
            assert_never(never)
    async with timeout_td(timeout):  # skipif-ci-and-not-linux
        response = await redis.publish(channel, data_use)  # skipif-ci-and-not-linux
    return ensure_int(response)  # skipif-ci-and-not-linux


@dataclass(kw_only=True, slots=True)
class PublishError(Exception):
    data: Any

    @override
    def __str__(self) -> str:
        return f"Unable to publish data {self.data!r} with no serializer"


##


async def publish_many[T](
    redis: Redis,
    channel: str,
    data: MaybeSequence[bytes | str | T],
    /,
    *,
    serializer: Callable[[T], EncodableT] | None = None,
    timeout: Delta | None = _PUBLISH_TIMEOUT,
) -> Sequence[bool]:
    """Publish an object/multiple objects to a channel."""
    async with TaskGroup() as tg:
        tasks = [
            tg.create_task(
                _try_publish(
                    redis,
                    channel,
                    d,
                    serializer=cast("Callable[[Any], EncodableT]", serializer),
                    timeout=timeout,
                )
            )
            for d in always_iterable(data)
        ]
    return [t.result() for t in tasks]


async def _try_publish[T](
    redis: Redis,
    channel: str,
    data: bytes | str | T,
    /,
    *,
    serializer: Callable[[T], EncodableT] | None = None,
    timeout: Delta | None = _PUBLISH_TIMEOUT,
) -> bool:
    try:
        _ = await publish(redis, channel, data, serializer=serializer, timeout=timeout)
    except TimeoutError:
        return False
    return True


##


_SUBSCRIBE_TIMEOUT: Delta = SECOND
_SUBSCRIBE_SLEEP: Delta = MILLISECOND


@overload
@enhanced_async_context_manager
def subscribe(
    redis: Redis,
    channels: MaybeIterable[str],
    queue: Queue[_RedisMessage],
    /,
    *,
    timeout: Delta | None = _SUBSCRIBE_TIMEOUT,
    sleep: Delta = _SUBSCRIBE_SLEEP,
    output: Literal["raw"],
    filter_: Callable[[_RedisMessage], bool] | None = None,
) -> AsyncIterator[Task[None]]: ...
@overload
@enhanced_async_context_manager
def subscribe(
    redis: Redis,
    channels: MaybeIterable[str],
    queue: Queue[bytes],
    /,
    *,
    timeout: Delta | None = _SUBSCRIBE_TIMEOUT,
    sleep: Delta = _SUBSCRIBE_SLEEP,
    output: Literal["bytes"],
    filter_: Callable[[bytes], bool] | None = None,
) -> AsyncIterator[Task[None]]: ...
@overload
@enhanced_async_context_manager
def subscribe(
    redis: Redis,
    channels: MaybeIterable[str],
    queue: Queue[str],
    /,
    *,
    timeout: Delta | None = _SUBSCRIBE_TIMEOUT,
    sleep: Delta = _SUBSCRIBE_SLEEP,
    output: Literal["text"] = "text",
    filter_: Callable[[str], bool] | None = None,
) -> AsyncIterator[Task[None]]: ...
@overload
@enhanced_async_context_manager
def subscribe[T](
    redis: Redis,
    channels: MaybeIterable[str],
    queue: Queue[T],
    /,
    *,
    timeout: Delta | None = _SUBSCRIBE_TIMEOUT,
    sleep: Delta = _SUBSCRIBE_SLEEP,
    output: Callable[[bytes], T],
    filter_: Callable[[T], bool] | None = None,
) -> AsyncIterator[Task[None]]: ...
@enhanced_async_context_manager
async def subscribe[T](
    redis: Redis,
    channels: MaybeIterable[str],
    queue: Queue[_RedisMessage] | Queue[bytes] | Queue[T],
    /,
    *,
    timeout: Delta | None = _SUBSCRIBE_TIMEOUT,
    sleep: Delta = _SUBSCRIBE_SLEEP,
    output: Literal["raw", "bytes", "text"] | Callable[[bytes], T] = "text",
    filter_: Callable[[Any], bool] | None = None,
) -> AsyncIterator[Task[None]]:
    """Subscribe to the data of a given channel(s)."""
    channels = list(always_iterable(channels))  # skipif-ci-and-not-linux
    match output:  # skipif-ci-and-not-linux
        case "raw":
            transform = cast("Any", identity)
        case "bytes":
            transform = cast("Any", itemgetter("data"))
        case "text":

            def transform(message: _RedisMessage, /) -> str:  # pyright: ignore[reportRedeclaration]
                return message["data"].decode()

        case Callable() as deserialize:

            def transform(message: _RedisMessage, /) -> T:
                return deserialize(message["data"])

        case never:
            assert_never(never)

    task = create_task(  # skipif-ci-and-not-linux
        _subscribe_core(
            redis,
            channels,
            transform,
            queue,
            timeout=timeout,
            sleep=sleep,
            filter_=filter_,
        )
    )
    try:  # skipif-ci-and-not-linux
        yield task
    finally:  # skipif-ci-and-not-linux
        try:
            _ = task.cancel()
        except RuntimeError as error:  # pragma: no cover
            if (not is_pytest()) or (error.args[0] != "Event loop is closed"):
                raise
        with suppress(CancelledError):
            await task


async def _subscribe_core(
    redis: Redis,
    channels: MaybeIterable[str],
    transform: Callable[[_RedisMessage], Any],
    queue: Queue[Any],
    /,
    *,
    timeout: Delta | None = _SUBSCRIBE_TIMEOUT,
    sleep: Delta = _SUBSCRIBE_SLEEP,
    filter_: Callable[[Any], bool] | None = None,
) -> None:
    timeout_use = (  # skipif-ci-and-not-linux
        None if timeout is None else to_seconds(timeout)
    )
    is_subscribe_message = partial(  # skipif-ci-and-not-linux
        _is_message, channels={c.encode() for c in channels}
    )
    async with yield_pubsub(redis, channels) as pubsub:  # skipif-ci-and-not-linux
        while True:
            message = await pubsub.get_message(timeout=timeout_use)
            if is_subscribe_message(message):
                transformed = transform(message)
                if (filter_ is None) or filter_(transformed):
                    queue.put_nowait(transformed)
            else:
                await sleep_td(sleep)


def _is_message(
    message: Any, /, *, channels: Collection[bytes]
) -> TypeGuard[_RedisMessage]:
    return (
        isinstance(message, Mapping)
        and ("type" in message)
        and (message["type"] in {"subscribe", "psubscribe", "message", "pmessage"})
        and ("pattern" in message)
        and ((message["pattern"] is None) or isinstance(message["pattern"], str))
        and ("channel" in message)
        and (message["channel"] in channels)
        and ("data" in message)
        and isinstance(message["data"], bytes)
    )


class _RedisMessage(TypedDict):
    type: Literal["subscribe", "psubscribe", "message", "pmessage"]
    pattern: str | None
    channel: bytes
    data: bytes


##


@enhanced_async_context_manager
async def yield_pubsub(
    redis: Redis, channels: MaybeIterable[str], /
) -> AsyncIterator[PubSub]:
    """Yield a PubSub instance subscribed to some channels."""
    pubsub = redis.pubsub()  # skipif-ci-and-not-linux
    channels = list(always_iterable(channels))  # skipif-ci-and-not-linux
    await pubsub.subscribe(*channels)  # skipif-ci-and-not-linux
    try:  # skipif-ci-and-not-linux
        yield pubsub
    finally:  # skipif-ci-and-not-linux
        await pubsub.unsubscribe(*channels)
        await pubsub.aclose()


##


_HOST = "localhost"
_PORT = 6379


@enhanced_async_context_manager
async def yield_redis(
    *,
    host: str = _HOST,
    port: int = _PORT,
    db: str | int = 0,
    password: str | None = None,
    socket_timeout: float | None = None,
    socket_connect_timeout: float | None = None,
    socket_keepalive: bool | None = None,
    socket_keepalive_options: Mapping[int, int | bytes] | None = None,
    connection_pool: ConnectionPool | None = None,
    decode_responses: bool = False,
    **kwargs: Any,
) -> AsyncIterator[Redis]:
    """Yield an asynchronous redis client."""
    redis = Redis(
        host=host,
        port=port,
        db=db,
        password=password,
        socket_timeout=socket_timeout,
        socket_connect_timeout=socket_connect_timeout,
        socket_keepalive=socket_keepalive,
        socket_keepalive_options=socket_keepalive_options,
        connection_pool=connection_pool,
        decode_responses=decode_responses,
        **kwargs,
    )
    try:
        yield redis
    finally:
        await redis.aclose()


##


def _serialize[T](
    obj: T, /, *, serializer: Callable[[T], bytes] | None = None
) -> bytes:
    if serializer is None:  # skipif-ci-and-not-linux
        from utilities.orjson import serialize as serializer_use
    else:  # skipif-ci-and-not-linux
        serializer_use = serializer
    return serializer_use(obj)  # skipif-ci-and-not-linux


def _deserialize[T](
    data: bytes, /, *, deserializer: Callable[[bytes], T] | None = None
) -> T:
    if deserializer is None:  # skipif-ci-and-not-linux
        from utilities.orjson import deserialize as deserializer_use
    else:  # skipif-ci-and-not-linux
        deserializer_use = deserializer
    return deserializer_use(data)  # skipif-ci-and-not-linux


__all__ = [
    "RedisHashMapKey",
    "RedisKey",
    "publish",
    "publish_many",
    "redis_hash_map_key",
    "redis_key",
    "subscribe",
    "yield_pubsub",
    "yield_redis",
]



from __future__ import annotations

from contextlib import AbstractContextManager, suppress
from logging import LogRecord, setLogRecordFactory
from os import environ
from typing import TYPE_CHECKING

from hypothesis import HealthCheck
from pytest import fixture, mark, param, skip

from utilities.contextlib import enhanced_context_manager
from utilities.platform import IS_LINUX, IS_MAC, IS_NOT_LINUX, IS_WINDOWS
from utilities.re import ExtractGroupError, extract_group
from utilities.whenever import MINUTE, get_now, parse_plain_local

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Iterator, Sequence
    from pathlib import Path

    from _pytest.fixtures import SubRequest
    from redis.asyncio import Redis
    from sqlalchemy import Engine, TextClause
    from sqlalchemy.ext.asyncio import AsyncEngine


FLAKY = mark.flaky(reruns=5, reruns_delay=1)
IS_CI = "CI" in environ
SKIPIF_CI = mark.skipif(IS_CI, reason="Skipped for CI")
IS_CI_AND_NOT_LINUX = IS_CI and IS_NOT_LINUX
IS_CI_AND_WINDOWS = IS_CI and IS_WINDOWS
IS_CI_AND_MAC = IS_CI and IS_MAC
IS_CI_AND_LINUX = IS_CI and IS_LINUX
SKIPIF_CI_AND_NOT_LINUX = mark.skipif(
    IS_CI_AND_NOT_LINUX, reason="Skipped for CI/non-Linux"
)
SKIPIF_CI_AND_WINDOWS = mark.skipif(IS_CI_AND_WINDOWS, reason="Skipped for CI/Windows")
SKIPIF_CI_AND_MAC = mark.skipif(IS_CI_AND_MAC, reason="Skipped for CI/Mac")
SKIPIF_CI_AND_LINUX = mark.skipif(IS_CI_AND_LINUX, reason="Skipped for CI/Linux")


# hypothesis


try:
    from utilities.hypothesis import setup_hypothesis_profiles
except ModuleNotFoundError:
    pass
else:
    setup_hypothesis_profiles(suppress_health_check={HealthCheck.differing_executors})


# fixture - logging


@fixture
def set_log_factory() -> AbstractContextManager[None]:
    @enhanced_context_manager
    def cm() -> Iterator[None]:
        try:
            yield
        finally:
            setLogRecordFactory(LogRecord)

    return cm()


# fixtures - redis


@fixture
async def test_redis() -> AsyncIterator[Redis]:
    if IS_CI_AND_NOT_LINUX:
        skip(reason="Skipped for CI/non-Linux")

    from utilities.redis import yield_redis

    async with yield_redis(db=15) as redis:
        yield redis


# fixtures - sqlalchemy


@fixture(params=[param("sqlite"), param("postgresql", marks=SKIPIF_CI)])
def test_engine(*, request: SubRequest, tmp_path: Path) -> Engine:
    from sqlalchemy.exc import OperationalError

    from utilities.sqlalchemy import create_engine

    dialect = request.param
    match dialect:
        case "sqlite":
            db_path = tmp_path / "db.sqlite"
            return create_engine("sqlite", database=str(db_path))
        case "postgresql":
            engine = create_engine(
                "postgresql+psycopg", host="localhost", port=5432, database="testing"
            )
            try:
                with engine.begin() as conn:
                    tables: Sequence[str] = (
                        conn.execute(_select_tables()).scalars().all()
                    )
            except OperationalError:
                ...
            else:
                for table in filter(_is_to_drop, tables):
                    with engine.begin() as conn, suppress(Exception):
                        _ = conn.execute(_drop_table(table))
            return engine
        case _:
            msg = f"Unsupported dialect: {dialect}"
            raise NotImplementedError(msg)


@fixture(params=[param("sqlite"), param("postgresql", marks=SKIPIF_CI)])
async def test_async_engine(
    *,
    request: SubRequest,
    test_async_sqlite_engine: AsyncEngine,
    test_async_postgres_engine: AsyncEngine,
) -> AsyncEngine:
    dialect = request.param
    match dialect:
        case "sqlite":
            return test_async_sqlite_engine
        case "postgresql":
            return test_async_postgres_engine
        case _:
            msg = f"Unsupported dialect: {dialect}"
            raise NotImplementedError(msg)


@fixture
async def test_async_sqlite_engine(*, tmp_path: Path) -> AsyncEngine:
    from utilities.sqlalchemy import create_engine

    db_path = tmp_path / "db.sqlite"
    return create_engine("sqlite+aiosqlite", database=str(db_path), async_=True)


@fixture
async def test_async_postgres_engine() -> AsyncEngine:
    from asyncpg.exceptions import InvalidCatalogNameError

    from utilities.sqlalchemy import create_engine

    if IS_CI:
        skip(reason="Skipped for CI")
    engine = create_engine(
        "postgresql+asyncpg",
        host="localhost",
        port=5432,
        database="testing",
        async_=True,
    )
    try:
        async with engine.begin() as conn:
            tables: Sequence[str] = (
                (await conn.execute(_select_tables())).scalars().all()
            )
    except InvalidCatalogNameError:
        ...
    else:
        for table in filter(_is_to_drop, tables):
            async with engine.begin() as conn:
                with suppress(Exception):
                    _ = await conn.execute(_drop_table(table))
    return engine


def _is_to_drop(table: str, /) -> bool:
    try:
        datetime_str = extract_group(r"^(\d{8}T\d{6})_", table)
    except ExtractGroupError:
        return True
    return (get_now() - parse_plain_local(datetime_str)) >= MINUTE


def _select_tables() -> TextClause:
    from sqlalchemy import text

    return text("SELECT tablename FROM pg_tables")


def _drop_table(table: str, /) -> TextClause:
    from sqlalchemy import text

    return text(f'DROP TABLE IF EXISTS "{table}" CASCADE')



from __future__ import annotations

from re import search
from typing import TYPE_CHECKING, ClassVar, Literal

from tests.conftest import SKIPIF_CI
from utilities.asyncio import sleep_td
from utilities.fastapi import yield_ping_receiver
from utilities.whenever import SECOND

if TYPE_CHECKING:
    from whenever import TimeDelta


class TestPingReceiver:
    delta: ClassVar[TimeDelta] = 0.1 * SECOND
    port: ClassVar[int] = 5465

    @SKIPIF_CI
    async def test_main(self) -> None:
        assert await self.ping() is False
        await sleep_td(self.delta)
        async with yield_ping_receiver(self.port, timeout=2 * self.delta):
            await sleep_td(self.delta)
            result = await self.ping()
            assert isinstance(result, str)
            assert search(
                r"pong @ \d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{1,6}", result
            )
        await sleep_td(self.delta)
        assert await self.ping() is False

    async def ping(self) -> str | Literal[False]:
        """Ping the receiver."""
        from httpx import AsyncClient, ConnectError  # skipif-ci

        url = f"http://localhost:{self.port}/ping"  # skipif-ci
        try:  # skipif-ci
            async with AsyncClient() as client:
                response = await client.get(url)
        except ConnectError:  # skipif-ci
            return False
        return response.text if response.status_code == 200 else False  # skipif-ci
